/*
Definitie garpQuantityDefinition class
*/

:-pce_begin_class(
		  garpQuantityDefinition(name),
		  object,
		  "garp quantity definition"
		 ).

variable(name,name,get,"Name of the quantity definition").
variable(remarks,string,none,"Remarks on this quantity definition").

initialise(QD,
	Name : name,
	AllowedQS : chain,
	Remarks : [string]
	):->
	%initialiseer de definitie, zodat ie goed past op CR's..

	QD->+initialise,
	QD->>name(Name),
	default(Remarks,new(string),RRemarks),
	QD->>remarks(RRemarks),
	AllowedQS->>for_all(->>(QD,hyper,@arg1,allowedQuantitySpace,quantityDefinition)).
%%

%%
name(QD,
     N : name
    ):->
	"Set the name" ::

	QD->>slot(name,N?makeGarp).
%%

%%
quantities(QD,
			Quantities: chain
			):<-
	Quantities = QD<<-all_named_hypered(garpQuantity).
%%

%%
remarks(QD,
	Remarks : remarks = string
       ):->
	"Set a copy of the given string as remarks" ::

	QD->>slot(remarks,?(Remarks,strip,both)).
%%

%%
remarks(QD,
	Remarks : string
       ):<-
	"Return a copy of the remarks" ::

	RealRemarks = QD<<-slot(remarks),
	Remarks = RealRemarks<<-copy.
%%

%%
quantitySpaceAllowed(QD,
	QS : quantitySpace
	):->
	"Succeeds if the given qs is allowed with this definition" ::

	QD<<-hypered(allowedQuantitySpace,
					@arg3 == QS).
%%

%%
allowedQuantitySpaces(QD,
	C : chain
	):<-
	"Returns name-sorted chain of all quantity spaces that are allowed with this definition" ::

	C = QD<<-all_named_hypered(allowedQuantitySpace),
	C->>sort(?(@arg1?name,compare,@arg2?name)).
%%

%%%%%%%%CHANGES%%%%%%%%%%%%%%%%%%%%%%
checkChange_changeQuantityDef(QD,
			    CR : changeRequestor
			   ):->
	%als we zelf het object van de change zijn checken we of
	%alle data correct is (via een norm methode bij het model)
	%als we niet zelf het object vd change zijn checken we op
	%naamconflict..

	%1: we zijn het object

	(   CR->>checkObject(QD)
	->  @model->>norm_quantityDef(CR,QD)
	;   NewName = CR<<-argument(1),
		NewName->>equalExportName(QD?name), %gelijk?
	    CR->>impossible('The given name is already used',QD)
	).
%
applyChange_changeQuantityDef(QD,
			      CR: changeRequestor
			     ):->
	%Deze definitie verandert dus
	
	NewName = CR<<-argument(1),
	QD->>name(NewName),
	AllowedQS = CR<<-argument(2),
	%oude hypers weg
	QD->>delete_hypers(allowedQuantitySpace),
	AllowedQS->>for_all(->>(QD,hyper,@arg1,allowedQuantitySpace,quantityDefinition)),
	NewRemarks = CR<<-argument(3),
	QD->>remarks(NewRemarks).
%%

%%
checkChange_addQuantityDef(QD,
	CR:changeRequestor):->
	%checken op naamconflict

	NewName = CR<<-argument(1),
	NewName->>equalExportName(QD?name),
	CR->>impossible('The given name is already used', QD).
%%

%%
checkChange_deleteQuantitySpace(QD,
	CR:changeRequestor):->
	%checken of dit een toegestane QS is bij deze definitie.
	%als dat zo is hangt het er nog van af of er nog QS-en over zijn na de verwijdering..

	QS = CR<<-argument(1),
	if
		QD->>quantitySpaceAllowed(QS)
	then (
		if
			1 = QD?allowedQuantitySpaces<<-size
		then
			%ok, dan mag het gewoon niet
			CR->>impossible(string('The quantity definition "%s" has no other allowed quantity spaces',
									QD?name),
							QD)
		else
			%mag wel, doch wel waarschuwen
			CR->>warning(
				string('The quantity space will be removed from the list of the quantity definition "%s"',
							QD?name),
				QD)
		).
%%

		
:-pce_end_class.
		  
