/*
Definitie quantitySpace class

2000 - 2006 Jelmer Jellema, spininhetweb.nl
*/

:-pce_begin_class(
		  quantitySpace(name),
		  object,
		  "quantitySpace definition"
		 ).

:-pce_group('Model data').

variable(name,name,get,"Name of the quantitySpace").
variable(values,chain,none,"List of all values in the quantitySpace"). 
%de eerste in de chain is de hoogste
variable(remarks,string,none,"Remarks on this quantity definition").


:-pce_group(@default).
initialise(QS, Name: name = name,
			Values: chain,
			Remarks: [string]) :->
	send(QS,send_super,initialise),
	QS->>name(Name),
	QS->>slot(values,Values?copy),
	default(Remarks,new(string),RRemarks),
	QS->>remarks(RRemarks).
%%

%%
name(QS,
     N : name
    ):->
	"Set the name" ::

	QS->>slot(name,N?makeGarp).
%%

%%
remarks(QD,
	Remarks : remarks = string
       ):->
	"Set a copy of the given string as remarks" ::

	QD->>slot(remarks,?(Remarks,strip,both)).
%%

%%
remarks(QD,
	Remarks : string
       ):<-
	"Return a copy of the remarks" ::

	RealRemarks = QD<<-slot(remarks),
	Remarks = RealRemarks<<-copy.
%%

%%
values(QD,
	V : chain):<-
	"Return a new list containing copies of the values (with same ID)" ::

	RV = QD<<-slot(values),
	V = RV<<-map(@arg1?copy).
%%

%%%%%%%%CHANGES%%%%%%%%%%%%%%%%%%%%%%
checkChange_changeQuantitySpace(QS,
			    CR : changeRequestor
			   ):->
	%als we zelf het object van de change zijn checken we of
	%alle data correct is.
	%als we niet zelf het object vd change zijn checken we op
	%naamconflict..

	%1: we zijn het object: de norm staat bij het model (ivm addQuantitySpace)
	CR->>checkObject(QS),!,
	@model->>norm_quantitySpace(CR,QS).
%
checkChange_changeQuantitySpace(QS,CR):->
	%2: we zijn niet het object
	
	NewName = CR<<-argument(1),
	NewName->>equalExportName(QS?name), %gelijk?
	CR->>impossible('The given name is already used',QS).
%
applyChange_changeQuantitySpace(QS,
			      CR: changeRequestor
			     ):->
	%Deze qs verandert dus
	
	NewName = CR<<-argument(1),
	QS->>name(NewName),
	NewValues = CR<<-argument(2),
	RealValues = NewValues<<-map(@arg1?copy),
	QS->>slot(values,RealValues), %dus een kopie variant
	NewRemarks = CR<<-argument(3),
	QS->>remarks(NewRemarks).
%%

%%
checkChange_addQuantitySpace(QS,CR:changeRequestor):->
	%naam conflict checken dus
	NewName = CR<<-argument(1),
	NewName->>equalExportName(QS?name),
	CR->>impossible('The given name is already used',QS).
%%

%%%helpers

%%%%%%%%%%%%EXPORT%%%%%%%%%%%%%%%%%%
export(QS,F: file):->
	%export deze qs

	if
		(\+ 0 = QS?remarks<<-size)
	then
		F->>format('\n%s\n',?(string('%s\n%s\n', QS?name,QS?remarks),makeRemarks,0))
	else
		F->>format('\n%%%s\n',QS?name),
		
	%is er een punt dat niet zero is
	if
		QS?values->>find(and(
			@arg1?type == point,
			@arg1?valueName \== 'Zero'))
	then
		ContextVar = 'X'
	else
		ContextVar = '_',
		
	%
	EN = QS?name<<-exportName,
	F->>format('quantity_space(%s,%s,[\n'
				,EN,ContextVar),
				
	%waarden, die willen we van laag naar hoog
	Reversed *= chain,
	QS?values->>for_all(
		->>(Reversed,prepend,@arg1)),
	
	Comma *= var,
	Comma->>assign('',global),
	Reversed->>for_all(
		and(
			->>(F,format,'%s\t',Comma),
			assign(Comma,',\n',global),
			if(@arg1?valueName == 'Zero',
				->>(F,append,'point(zero)'),
				if(@arg1?type == interval,
					->>(F,append,@arg1?valueName?exportName),
					->>(F,format,'point(%s(X))',
						@arg1?valueName?exportName)
				)
			)
		)
	),
	
	F->>append('\n]).\n').
%%

:-pce_end_class.
		  
